from emitter import Client
import Tkinter
import json

emitter = Client()

root = Tkinter.Tk()

channel_key = tkinter.StringVar(root, value="8jMLP9F859oDyqmJ3aV4aqnmFZpxApvb")
#channel_key = tkinter.StringVar(root, value="aghbt67CuPawxQvoBfKZ8MpecpPoz7od")#local
channel = tkinter.StringVar(root, value="test/")
shortcut = tkinter.StringVar(root, value="a0")
text_message = tkinter.StringVar(root, value="Hello World")
share_group = tkinter.StringVar(root, value="sg")
share_group_key = tkinter.StringVar(root, value="b7FEsiGFQoSYA6qyeu1dDodFnO0ypp0f")
#share_group_key = tkinter.StringVar(root, value="Q_dM5ODuhWjaR_LNo886hVjoecvt5pMJ") #local

def connect():
	#emitter.connect(host="127.0.0.1", port=8080, secure=False)
	emitter.connect()

	emitter.on_connect = lambda: result_text.insert("0.0", "Connected\n\n")
	emitter.on_disconnect = lambda: result_text.insert("0.0", "Disconnected\n\n")
	emitter.on_presence = lambda p: result_text.insert("0.0", "Presence message: '" + p.as_string() + "'\n\n")
	emitter.on_message = lambda m: result_text.insert("0.0", "Message received on default handler, destined to " + m.channel + ": " + m.as_string() + "\n\n")
	emitter.on_error = lambda e: result_text.insert("0.0", "Error received: " + str(e) + "\n\n")
	emitter.on_me = lambda me: result_text.insert("0.0", "Information about Me received: " + str(me) +"\n\n")
	emitter.loop_start()

def disconnect():
	emitter.loop_stop()
	emitter.disconnect()

def subscribe(share_group=None):
	str_key = channel_key.get()
	str_channel = channel.get()
	emitter.subscribe(str_key,
	 str_channel,
	 optional_handler=lambda m: result_text.insert("0.0", "Message received on handler for " + str_channel + ": " + m.as_string() + "\n\n"))
	result_text.insert("0.0", "Subscribtion to '" + str_channel + "' requested.\n\n")

def subscribe_share():
	str_key = share_group_key.get()
	str_channel = channel.get()
	str_share = share_group.get()
	emitter.subscribe_with_group(str_key,
	 str_channel,
	 optional_handler=lambda m: result_text.insert("0.0", "Message received on handler for " + str_channel + ": " + m.as_string() + "\n\n"),
	 share_group=str_share)
	result_text.insert("0.0", "Subscribtion to '" + str_channel + "' requested.\n\n")

def unsubscribe():
	str_key = channel_key.get()
	str_channel = channel.get()
	emitter.unsubscribe(str_key, str_channel)
	result_text.insert("0.0", "Unsubscribtion to '" + str_channel + "' requested.\n\n")

def presence():
	str_key = channel_key.get()
	str_channel = channel.get()
	emitter.presence(str_key, str_channel, status=True, changes=True)
	result_text.insert("0.0", "Presence on '" + str_channel + "' requested.\n\n")

def message(retain=False):
	if retain:
		emitter.publish(channel_key.get(), channel.get(), text_message.get(), {Client.with_retain()})
	else:
		emitter.publish(channel_key.get(), channel.get(), text_message.get(), {})
	result_text.insert("0.0", "Test message send through '" + channel.get() + "'.\n\n")

def link():
	str_key = channel_key.get()
	str_channel = channel.get()
	str_link = shortcut.get()
	emitter.link(str_key, str_channel, str_link, False, True)

def link_private():
	str_key = channel_key.get()
	str_channel = channel.get()
	str_link = shortcut.get()
	emitter.link(str_key, str_channel, str_link, True, True)

def pub_to_link():
	str_link = shortcut.get()
	emitter.publish_with_link(str_link, text_message.get())

def me():
	emitter.me()

# Col 1
Tkinter.Label(root, text="Channel : ").grid(column=1, row=1)
channel_entry = Tkinter.Entry(root, width=40, textvariable=channel)
channel_entry.grid(column=1, row=2)

Tkinter.Label(root, text="Channel key : ").grid(column=1, row=3)
channel_key_entry = Tkinter.Entry(root, width=40, textvariable=channel_key)
channel_key_entry.grid(column=1, row=4)

Tkinter.Label(root, text="Shortcut : ").grid(column=1, row=5)
shortcut_entry = Tkinter.Entry(root, width=40, textvariable=shortcut)
shortcut_entry.grid(column=1, row=6)

Tkinter.Label(root, text="Message : ").grid(column=1, row=7)
message_entry = Tkinter.Entry(root, width=40, textvariable=text_message)
message_entry.grid(column=1, row=8)

Tkinter.Label(root, text="Share group : ").grid(column=1, row=9)
share_entry = Tkinter.Entry(root, width=40, textvariable=share_group)
share_entry.grid(column=1, row=10)

Tkinter.Label(root, text="Share group key : ").grid(column=1, row=11)
share_key_entry = Tkinter.Entry(root, width=40, textvariable=share_group_key)
share_key_entry.grid(column=1, row=12)


# Col 2
connect_button = Tkinter.Button(root, text="Connect", width=30, command=connect)
connect_button.grid(column=2, row=1)

disconnect_button = Tkinter.Button(root, text="Disconnect", width=30, command=disconnect)
disconnect_button.grid(column=2, row=2)

subscribe_button = Tkinter.Button(root, text="Subscribe", width=30, command=subscribe)
subscribe_button.grid(column=2, row=4)

unsubscribe_button = Tkinter.Button(root, text="Unsubscribe", width=30, command=unsubscribe)
unsubscribe_button.grid(column=2, row=5)

subscribe_share_button = Tkinter.Button(root, text="Subscribe to share", width=30, command=subscribe_share)
subscribe_share_button.grid(column=2, row=6)

# Col 3
link_button = Tkinter.Button(root, text="Link to shortcut", width=30, command=link)
link_button.grid(column=3, row=1)

link_private_button = Tkinter.Button(root, text="Link to private channel", width=30, command=link_private)
link_private_button.grid(column=3, row=2)


send_button = Tkinter.Button(root, text="Publish to channel", width=30, command=message)
send_button.grid(column=3, row=4)

send_button = Tkinter.Button(root, text="Publish to channel with retain", width=30, command=lambda: message(retain=True))
send_button.grid(column=3, row=5)

pub_to_link_button = Tkinter.Button(root, text="Publish to link", width=30, command=pub_to_link)
pub_to_link_button.grid(column=3, row=6)

# Col 4
presence_button = Tkinter.Button(root, text="Presence", width=30, command=presence)
presence_button.grid(column=4, row=1)

me_button = Tkinter.Button(root, text="Me", width=30, command=me)
me_button.grid(column=4, row=2)

# Text area
result_text = Tkinter.Text(root, height=30, width=120)
result_text.grid(column=1, row=14, columnspan=4)

root.mainloop()


import pytest
from subtrie import SubTrie


def test_lookup_with_wildcard():
    t = SubTrie()
    t.insert("a/", lambda: None)
    t.insert("a/b/c/", lambda: None)
    t.insert("a/+/c/", lambda: None)
    t.insert("a/b/c/d/", lambda: None)
    t.insert("a/+/c/+/", lambda: None)
    t.insert("x/", lambda: None)
    t.insert("x/y/", lambda: None)
    t.insert("x/+/z/", lambda: None)

    tests =	{
        "a/": 1,
		"a/1/": 1,
		"a/2/": 1,
		"a/1/2/": 1,
		"a/1/2/3/": 1,
		"a/x/y/c/": 1,
		"a/x/c/": 2,
		"a/b/c/": 3,
		"a/b/c/d/": 5,
		"a/b/c/e/": 4,
		"x/y/c/e/": 2}

    for k, v in tests.items():
        results = t.lookup(k)
        assert len(results) == v


def test_delete_parent():
    t = SubTrie()
    t.insert("a", lambda: None)
    t.insert("a/b/c", lambda: None)
    t.insert("a/+/c", lambda: None)

    t.delete("a")

    results = t.lookup("a")
    assert len(results) == 0

    results = t.lookup("a/b")
    assert len(results) == 0
    
    results = t.lookup("a/b/c")
    assert len(results) == 2

def test_delete_child():
    t = SubTrie()
    t.insert("a", lambda: None)
    t.insert("a/b", lambda: None)

    results = t.lookup("a/b")
    assert len(results) == 2

    t.delete("a/b")

    results = t.lookup("a/b")
    assert len(results) == 1

def test_delete_inexistent_child():
    t = SubTrie()
    t.insert("a", lambda: None)
    t.insert("a/b", lambda: None)

    t.delete("c")

    results = t.lookup("a/b")
    assert len(results) == 2


def test_delete_root():
    t = SubTrie()
    t.insert("a", lambda: None)

    t.delete("a")

    results = t.lookup("a")
    assert len(results) == 0


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="emitter-io",
    version="2.0.0",
    author="Florimond Husquinet",
    author_email="florimond@emitter.io",
    description="A Python library to interact with the Emitter API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://emitter.io",
    packages=["emitter"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Communications",
        "Topic :: Internet :: WWW/HTTP",
        "License :: OSI Approved :: Eclipse Public License 1.0 (EPL-1.0)",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3"
    ],
    keywords="emitter mqtt realtime cloud service",
    install_requires=["paho-mqtt"]
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This is the Python client for Emitter (emitter.io).
GitHub: github.com/emitter-io/python
License: Eclipse Public License 1.0 (EPL-1.0)
"""
import json
import re
import logging
import ssl
import paho.mqtt.client as mqtt
from subtrie import SubTrie

class Client(object):
	"""
	* Represents the client connection to an Emitter server.
	"""

	def __init__(self):
		"""
		* Register the variables for later use.
		"""
		self._mqtt = None
		self._handler_message = None
		self._handler_trie_message = SubTrie()
		self._handler_connect = None
		self._handler_disconnect = None
		self._handler_error = None
		self._handler_presence = None
		self._handler_trie_presence = SubTrie()
		self._handler_me = None
		self._handler_keygen = None

	@property
	def on_connect(self):
		return self._handler_connect
	@on_connect.setter
	def on_connect(self, func):
		self._handler_connect = func

	@property
	def on_disconnect(self):
		return self._handler_disconnect
	@on_disconnect.setter
	def on_disconnect(self, func):
		self._handler_disconnect = func

	@property
	def on_error(self):
		return self._handler_error
	@on_error.setter
	def on_error(self, func):
		self._handler_error = func

	@property
	def on_presence(self):
		return self._handler_presence
	@on_presence.setter
	def on_presence(self, func):
		self._handler_presence = func

	@property
	def on_me(self):
		return self._handler_me
	@on_me.setter
	def on_me(self, func):
		self._handler_me = func

	@property
	def on_keygen(self):
		return self._handler_keygen
	@on_keygen.setter
	def on_keygen(self, func):
		self._handler_keygen = func

	@property
	def on_message(self):
		return self._handler_message
	@on_message.setter
	def on_message(self, func):
		self._handler_message = func

	def loop(self, timeout):
		"""
		* Call regularly to process network events. This call waits in select()
		* until the network socket is available for reading or writing, if
		* appropriate, then handles the incoming/outgoing data. This function
		* blocks for up to timeout seconds.
		* timeout must not exceed the keepalive value for the client or your
		* client will be regularly disconnected by the broker.
		"""
		self._mqtt.loop(timeout=timeout)

	def loop_forever(self):
		"""
		* This is a blocking form of the network loop and will not return until the
		* client calls disconnect(). It automatically handles reconnecting.
		"""
		self._mqtt.loop_forever()

	def loop_start(self):
		"""
		* These functions implement a threaded interface to the network loop.
		* Calling loop_start() once, before or after connect*(), runs a thread in
		* the background to call loop() automatically. This frees up the main
		* thread for other work that may be blocking. This call also handles
		* reconnecting to the broker.
		"""
		self._mqtt.loop_start()

	def loop_stop(self):
		"""
		* Stops the loop started in loopStart().
		* See loopStart() for more information.
		"""
		self._mqtt.loop_stop()


	def _on_connect(self, client, userdata, flags, rc):
		"""
		* Occurs when connection is established.
		"""
		if self._handler_connect:
			self._handler_connect()
		

	def _on_disconnect(self, client, userdata, rc):
		"""
		* Occurs when the connection was lost.
		"""
		if self._handler_disconnect:
			self._handler_disconnect()

	def _invoke_trie_handlers(self, trie, default_handler, message):
		handlers = trie.lookup(message.channel)
		if len(handlers) == 0 and default_handler:
				default_handler(message)

		for h in handlers:
			h(message)
		

	def _on_message(self, client, userdata, msg):
		message = EmitterMessage(msg)
		
		# Non-emitter messages are far more frequent, so if it is one, return earlier.
		if (not message.channel.startswith("emitter")):
			self._invoke_trie_handlers(self._handler_trie_message, self._handler_message, message)

		if self._handler_keygen and message.channel.startswith("emitter/keygen"):
			# This is a keygen message.
			self._handler_keygen(message.as_object())

		elif message.channel.startswith("emitter/presence"):
			# This is a presence message.
			self._invoke_trie_handlers(self._handler_trie_presence, self._handler_presence, message)

		elif self._handler_error and message.channel.startswith("emitter/error"):
			# This is an error message.
			self._handler_error(message.as_object())

		elif self._handler_me and message.channel.startswith("emitter/me"):
			# This is a "me" message, giving information about the connection.
			self._handler_me(message.as_object())
	
	def connect(self, host="api.emitter.io", port=443, secure=True, keepalive=30):
		"""
		* Connects to an Emitter server.
		"""
		formatted_host = re.sub(r"/.*?:\/\//g", "", host)
		self._mqtt = mqtt.Client()

		if secure:
			ssl_ctx = ssl.create_default_context()
			self._mqtt.tls_set_context(ssl_ctx)

		self._mqtt.on_connect = self._on_connect
		self._mqtt.on_disconnect = self._on_disconnect
		self._mqtt.on_message = self._on_message

		self._mqtt.connect(host=formatted_host, port=port, keepalive=keepalive)

	def publish(self, key, channel, message, options={}):
		"""
		* Publishes a message to a channel.
		"""
		topic = self._format_channel(key, channel, options)
		qos, retain = Client._get_header(options)

		self._mqtt.publish(topic, message, qos=qos, retain=retain)

	def subscribe(self, key, channel, optional_handler=None, options={}):
		"""
		* Subscribes to a particual share group.
		"""
		if optional_handler is not None:
			self._handler_trie_message.insert(channel, optional_handler)

		topic = Client._format_channel(key, channel, options)
		self._mqtt.subscribe(topic)

	def subscribe_with_group(self, key, channel, share_group, optional_handler=None, options={}):
		"""
		* Subscribes to a particual share group.
		"""
		if optional_handler is not None:
			self._handler_trie_message.insert(channel, optional_handler)

		topic = Client._format_channel_share(key, channel, share_group, options)
		self._mqtt.subscribe(topic)

	def unsubscribe(self, key, channel):
		"""
		* Unsubscribes from a particular channel.
		"""
		self._handler_trie_message.delete(channel)
		topic = self._format_channel(key, channel)
		self._mqtt.unsubscribe(topic)

	def disconnect(self):
		"""
		* Disconnects from the connected Emitter server.
		"""
		self._mqtt.disconnect()

	def presence(self, key, channel, status=False, changes=False, optional_handler=None):
		"""
		* Sends a presence request to the server.
		"""
		if optional_handler is not None:
			self._handler_trie_presence.insert(channel, optional_handler)

		request = {"key": key, "channel": channel, "status": status, "changes": changes}
		# Publish the request.
		self._mqtt.publish("emitter/presence/", json.dumps(request))

	def keygen(self, key, channel, permissions, ttl=0):
		"""
		* Sends a key generation request to the server.
		"""
		request = {"key": key, "channel": channel, "type": permissions, "ttl": ttl}
		# Publish the request.
		self._mqtt.publish("emitter/keygen/", json.dumps(request))

	def link(self, key, channel, name, private, subscribe, options={}):
		"""
		* Sends a link creation request to the server.
		"""
		formattedChannel = Client._format_channel_link(channel, options=options)
		request = {"key": key, "channel": formattedChannel, "name": name, "private": private, "subscribe": subscribe}

		# Publish the request.
		self._mqtt.publish("emitter/link/", json.dumps(request))

	def publish_with_link(self, link, message):
		"""
		* Sends a message through the link.
		"""
		# Publish the request.
		self._mqtt.publish(link, message)

	def me(self):
		"""
		* Requests information about the connection.
		"""
		self._mqtt.publish("emitter/me/", "")

	@staticmethod
	def _get_header(options):
		retain = False
		qos = 0
		for o in options:
			if o == Client.RETAIN:
				retain = True
			elif o == Client.QOS1:
				qos = 1
		
		return qos, retain

	@staticmethod
	def _format_options(options):
		formatted = ""

		if options != None and len(options) > 0:
			formatted = "?"
			
			for i, o in enumerate(options):
			
				if o.startswith("+"):
					continue
			
				formatted += o

				if i < len(options) - 1:
					formatted += "&"

		return formatted


	@staticmethod
	def _format_channel(key, channel, options={}):
		k = key.strip("/")
		c = channel.strip("/")
		o = Client._format_options(options)

		formatted = "{key}/{channel}/{options}".format(key=k, channel=c, options=o)
		return formatted

	@staticmethod
	def _format_channel_link(channel, options={}):
		c = channel.strip("/")
		o = Client._format_options(options)

		formatted = "{channel}/{options}".format(channel=c, options=o)
		return formatted

	@staticmethod
	def _format_channel_share(key, channel, share_group, options={}):
		k = key.strip("/")
		c = channel.strip("/")
		s = share_group.strip("/")
		o = Client._format_options(options)

		formatted = "{key}/$share/{share}/{channel}/{options}".format(key=k, share=s, channel=c, options=o)
		return formatted

	RETAIN = "+r"
	QOS0 = "+0"
	QOS1 = "+1"

	@staticmethod
	def with_ttl(ttl):
		return "ttl=" + str(ttl)

	@staticmethod
	def without_echo():
		return "me=0"

	@staticmethod
	def with_last(last):
		return "last=" + str(last)

	@staticmethod
	def with_retain():
		return Client.RETAIN
	
	@staticmethod
	def with_at_most_once():
		return Client.QOS0
	
	@staticmethod
	def with_at_least_once():
		return Client.QOS1

	

class EmitterMessage(object):
	"""
	* Represents a message received from the Emitter server.
	"""

	def __init__(self, message):
		"""
		* Creates an instance of EmitterMessage.
		"""
		self.channel = message.topic
		self.binary = message.payload

	def as_string(self):
		"""
		* Returns the payload as a utf-8 string.
		"""
		return str(self.binary)

	def as_object(self):
		"""
		* Returns the payload as an JSON-deserialized Python object.
		"""
		msg = None
		try:
			msg = json.loads(self.binary)
		except json.JSONDecodeError as exception:
			logging.exception(exception)

		return msg

	def as_binary(self):
		"""
		* Returns the payload as a raw binary buffer.
		"""
		return self.binary



class SubTrieNode:
    def __init__(self, parent, word, handler):
        self.parent = parent
        self.word = word
        self.children = {}
        self.handler = handler

class SubTrie:
    def __init__(self):
        self.root = SubTrieNode(None, None, None)

    @staticmethod
    def _get_words(topic):
        return filter(None, topic.split("/"))
    
    def insert(self, topic, handler):
        cur_node = self.root
        for word in self._get_words(topic):
            if word not in cur_node.children:
                cur_node.children[word] = SubTrieNode(cur_node, word, None)
            cur_node = cur_node.children[word]
        cur_node.handler = handler

    def _lookup(self, route, children):
        handlers = []

        if len(route) == 0:
            return handlers

        if route[0] in children:
            if children[route[0]].handler != None:
                handlers.append(children[route[0]].handler)
            handlers = handlers + self._lookup(route[1:], children[route[0]].children)
        
        if "+" in children:
            if children["+"].handler != None:
                handlers.append(children["+"].handler)
            handlers = handlers + self._lookup(route[1:], children["+"].children)
        
        return handlers

    def lookup(self, topic):
        route = list(self._get_words(topic))
        return self._lookup(route, self.root.children)
    
    def delete(self, topic):
        cur_node = self.root
        for word in self._get_words(topic):
            if word not in cur_node.children:
                return
            cur_node = cur_node.children[word]

        cur_node.handler = None
        
        while cur_node != self.root and cur_node.handler == None and len(cur_node.children) == 0:
            del cur_node.parent.children[cur_node.word]
            cur_node = cur_node.parent

        return


import pytest
import emitter

def test_formatChannel():
    tests = [
        {"key": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha", "channel": "test", "options": None, "expected": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/test/"},
        {"key": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha", "channel": "test/", "options": None, "expected": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/test/"},
        {"key": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/", "channel": "test", "options": None, "expected": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/test/"},
        {"key": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/", "channel": "test/", "options": None, "expected": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/test/"},
        # With options.
        {"key": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha", "channel": "test", "options": {"ttl":5}, "expected": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/test/?ttl=5"},
        {"key": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha", "channel": "test", "options": {"ttl":5,"me":0}, "expected": "5xZjIQp6GA9fpxso1Kslqnv8d4XVWCha/test/?ttl=5&me=0"},
        # Links (without a channel key).
        {"key": "", "channel": "test", "options": {"ttl":5,"me":0}, "expected": "test/?ttl=5&me=0"},
        {"key": None, "channel": "test", "options": {"ttl":5,"me":0}, "expected": "test/?ttl=5&me=0"},
        ]
    
    for test in tests:
        formatted = emitter.Emitter._formatChannel(key=test["key"], channel=test["channel"], options=test["options"])  
        assert formatted == test["expected"]
